import os
import json
from glob import glob
import subprocess
import shutil
import ast
import astor  # You need to install astor via pip
import tempfile
import re

# Set project root and configuration paths
PROJECT_ROOT = os.getcwd()

# Load the project configuration
config_path = os.path.join(PROJECT_ROOT, 'Dezyne', 'config.json')
if not os.path.exists(config_path):
    raise FileNotFoundError(f"Configuration file not found: {config_path}")

with open(config_path, 'r') as config_file:
    project_config = json.load(config_file)

PROJECT_NAME = project_config['projectName']
DEZYNE_PATH = project_config['dezynePath']
print(f"Project Name: {PROJECT_NAME}")

###########################################################
## Helper Functions                                      ##
###########################################################

def is_skel_component(file_path):
    """Check if a Dezyne file is a Skel file by checking if a component lacks behavior or system blocks."""
    with open(file_path, 'r') as f:
        content = f.read()

    # Look for a component declaration using a simple regex
    component_match = re.search(r'component\s+(\w+)\s*\{', content)
    if not component_match:
        return False  # No component found, not a Skel file

    # If a component is found, check if behavior or system block is present
    has_behavior = re.search(r'behavior\s*\{', content)
    has_system = re.search(r'system\s*\{', content)

    # If neither behavior nor system block exists, it's a Skel file
    return not (has_behavior or has_system)

def add_behavior_block(file_path):
    """Temporarily add a behavior block to a Skel file inside the component definition."""
    with open(file_path, 'r') as f:
        content = f.readlines()

    # Find the component's closing '}' and insert 'behavior {}' before it
    for i, line in enumerate(content):
        if line.strip() == '}':  # Found the closing bracket of the component
            content.insert(i, '    behavior {}\n')  # Insert behavior block before the closing '}'
            break

    # Write the modified content back to the file
    with open(file_path, 'w') as f:
        f.writelines(content)

def remove_behavior_block(file_path):
    """Remove the temporary behavior block from a Skel file."""
    with open(file_path, 'r') as f:
        content = f.readlines()

    # Remove the last line containing "behavior {}"
    with open(file_path, 'w') as f:
        for line in content:
            if line.strip() != "behavior {}":
                f.write(line)

def backup_file(file_path, backup_dir):
    """Backup a file to the specified directory."""
    backup_path = os.path.join(backup_dir, os.path.basename(file_path))
    shutil.copyfile(file_path, backup_path)
    return backup_path

def restore_file(file_path, backup_path):
    """Restore a file from its backup."""
    shutil.copyfile(backup_path, file_path)

def merge_python_classes(backup_file_path, new_file_path):
    """Merge classes from backup and new files, preserving custom methods."""
    
    with open(backup_file_path, 'r') as f:
        backup_content = f.read()
    with open(new_file_path, 'r') as f:
        new_content = f.read()

    backup_tree = ast.parse(backup_content)
    new_tree = ast.parse(new_content)

    backup_classes = {}
    new_classes = {}

    def extract_classes(tree, classes_dict):
        for node in tree.body:
            if isinstance(node, ast.ClassDef):
                methods = {item.name: item for item in node.body if isinstance(item, ast.FunctionDef)}
                classes_dict[node.name] = {'class_node': node, 'methods': methods}

    extract_classes(backup_tree, backup_classes)
    extract_classes(new_tree, new_classes)

    for class_name, new_class_info in new_classes.items():
        new_methods = new_class_info['methods']
        if class_name in backup_classes:
            backup_methods = backup_classes[class_name]['methods']
            merged_methods = [
                backup_methods.get(method_name, new_method_node) 
                for method_name, new_method_node in new_methods.items()
            ]
            new_class_info['class_node'].body = merged_methods

    merged_body = [
        node if isinstance(node, (ast.Import, ast.ImportFrom)) else new_classes.get(node.name, node)['class_node']
        for node in new_tree.body
    ]

    merged_tree = ast.Module(body=merged_body)
    with open(new_file_path, 'w') as f:
        f.write(astor.to_source(merged_tree))

###########################################################
## Dezyne Code Generation                                ##
###########################################################

# Find Dezyne models
dezyne_models = glob("Dezyne/Models/*.dzn")

# Detect Skel and normal models by checking their content
skel_models = [model for model in dezyne_models if is_skel_component(model)]
normal_models = [model for model in dezyne_models if not is_skel_component(model)]

# Use a temporary directory to store backups
with tempfile.TemporaryDirectory() as temp_backup_dir:
    print(f"Using temporary directory for backup: {temp_backup_dir}")

    # Handle Skel files
    if skel_models:
        skel_python_files = []
        for model in skel_models:
            model_name = os.path.splitext(os.path.basename(model))[0]
            model_py = os.path.join(PROJECT_ROOT, 'Dezyne', 'dzn', f"{model_name}.py")

            # Backup the original Skel file
            backup_path = backup_file(model, temp_backup_dir)
            
            # Add the behavior block before code generation
            add_behavior_block(model)

            if os.path.exists(model_py):
                backup_py = os.path.join(temp_backup_dir, f"{model_name}.py")
                shutil.copyfile(model_py, backup_py)
                skel_python_files.append((model_py, backup_py))
                print(f"Backed up {model_py} to {backup_py}")

    # Generate code for normal models
    if normal_models:
        print("Generating code for normal models...")
        for model in normal_models:
            model_name = os.path.splitext(os.path.basename(model))[0]
            model_py = os.path.join(PROJECT_ROOT, 'Dezyne', 'dzn', f"{model_name}.py")
            command = [os.path.join(DEZYNE_PATH, 'dzn'), "code", model, "-l", "python", "-o", os.path.join(PROJECT_ROOT, 'Dezyne', 'dzn')]
            try:
                subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                print(f"Successfully generated {model_py} from {model}")
            except subprocess.CalledProcessError as e:
                print(f"Error generating code for {model}: {e.stderr}")

    # Generate code for Skel models
    if skel_models:
        print("Generating code for Skel models...")
        for model in skel_models:
            model_name = os.path.splitext(os.path.basename(model))[0]
            model_py = os.path.join(PROJECT_ROOT, 'Dezyne', 'dzn', f"{model_name}.py")
            command = [os.path.join(DEZYNE_PATH, 'dzn'), "code", model, "-l", "python", "-o", os.path.join(PROJECT_ROOT, 'Dezyne', 'dzn')]
            try:
                subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                print(f"Successfully generated {model_py} from {model}")
            except subprocess.CalledProcessError as e:
                print(f"Error generating code for {model}: {e.stderr}")

            # After code generation, restore the original Skel file
            restore_file(model, backup_path)

    # Merge custom changes into Skel Python files
    if skel_models:
        for model_py, backup_py in skel_python_files:
            try:
                merge_python_classes(backup_py, model_py)
                print(f"Successfully merged custom changes into {model_py}")
            except Exception as e:
                print(f"Error during merging for {model_py}: {e}")

    print(f"Temporary backup directory {temp_backup_dir} has been removed.")
